from abc import ABC, abstractmethod
import inspect
import io
from typing import AsyncGenerator, Dict, Tuple, Type, Any
import aiohttp

FILE_CLASSES = (
    io.IOBase,
    aiohttp.StreamReader,
    AsyncGenerator,
)


class QueryVariable:
    @abstractmethod
    async def to_variable(self):
        raise NotImplementedError


async def parse_variables(
    variables: Dict, file_classes: Tuple[Type[Any], ...] = FILE_CLASSES
) -> Tuple[Dict, Dict]:
    files = {}

    async def recurse_extract(path, obj):
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """
        nonlocal files
        if isinstance(obj, list):
            nulled_obj = []
            for key, value in enumerate(obj):
                value = await recurse_extract(f"{path}.{key}", value)
                nulled_obj.append(value)
            return nulled_obj
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():
                value = await recurse_extract(f"{path}.{key}", value)
                nulled_obj[key] = value
            return nulled_obj
        elif isinstance(obj, file_classes):
            # extract obj from its parent and put it into files instead.
            files[path] = obj
            return None

        elif hasattr(obj, "to_variable"):
            return await obj.to_variable()

        else:
            # base case: pass through unchanged
            return obj

    nulled_variables = await recurse_extract("variables", variables)

    return nulled_variables, files



from abc import abstractproperty
import json
from typing import Any, Dict
from herre.exceptions import TokenExpired
from herre.wards.base import BaseQuery, BaseWard, WardException
import aiohttp
import logging
import re
from herre.wards.variables import parse_variables
from fakts import Fakts, get_current_fakts, Config

logger = logging.getLogger(__name__)

gqlparsed_with_variables = re.compile(
    r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)\((?P<arguments>[^\)]*)\)[\s]*{[\s]*(?P<firstchild>[^\(:]*).*"
)
gqlparser_without_variables = re.compile(
    r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)[\s]*{[\s]*(?P<firstchild>[^\(\{\s:]*).*"
)


class GQLException(Exception):
    pass


class ParsedQuery(BaseQuery):
    def __init__(self, query: str) -> None:
        self.query = query
        self.variables = None
        self._type = None
        self.m = gqlparsed_with_variables.match(self.query)
        self.has_variables = True

        if not self.m:
            self.m = gqlparser_without_variables.match(self.query)
            self.has_variables = False

        if not self.m:
            raise GQLException(f"Illformed request {self.query}")

    def get_operation_name_for_datakey(self, key):
        # TODO: Make work
        return key

    def combine(self, variables: dict):
        self.variables = variables
        return {"query": self.query, "variables": self.variables}

    def parsed(self):
        assert (
            self.variables and self.query
        ), "Please specify query and set variables before parsing"
        return {"query": self.query, "variables": self.variables}

    @property
    def firstchild(self):
        return self.m.group("firstchild")

    @property
    def operation_name(self):
        return self.m.group("operation")

    @property
    def type(self):
        return self.m.group("type")

    def extract(self, result: dict):
        assert (
            self.firstchild in result
        ), f"Cannot Access {self.firstchild}: in {result}"
        assert result[self.firstchild] is not None, f"Empty response {result}"
        return result[self.firstchild]


class GraphQLException(WardException):
    pass


class GraphQLQueryException(GraphQLException):
    pass


class GraphQLProtocolException(WardException):
    pass


class GraphQLWardConfig(Config):
    @abstractproperty
    def endpoint(self):
        return NotImplementedError("Please overwrite endpoint")


class GraphQLWard(BaseWard):
    configClass = GraphQLWardConfig
    config: GraphQLWardConfig

    class Meta:
        abstract = True

    async def handle_connect(self):

        self.async_session = aiohttp.ClientSession(headers=self.herre.headers)

    async def handle_disconnect(self):
        await self.async_session.close()
        self.connected = False

    async def handle_run(
        self, gql: ParsedQuery, variables: dict = {}, files: dict = {}, retry=0
    ):
        try:
            assert (
                retry < self.max_retries
            ), f"Retries Exceeded for Request {gql} with {variables}"

            payload = {"query": gql.query}

            if len(files.items()) > 0:
                # We have to uload files

                payload["variables"] = variables

                data = aiohttp.FormData()

                file_map = {str(i): [path] for i, path in enumerate(files)}
                # Enumerate the file streams
                # Will generate something like {'0': <_io.BufferedReader ...>}
                file_streams = {str(i): files[path] for i, path in enumerate(files)}
                operations_str = json.dumps(payload)

                data.add_field(
                    "operations", operations_str, content_type="application/json"
                )
                file_map_str = json.dumps(file_map)
                data.add_field("map", file_map_str, content_type="application/json")

                for k, v in file_streams.items():
                    data.add_field(
                        k,
                        v,
                        filename=getattr(v, "name", k),
                    )

                post_args: Dict[str, Any] = {"data": data}

            else:
                payload["variables"] = variables
                post_args = {"json": payload}

            async with self.async_session.post(
                self.config.endpoint, **post_args
            ) as resp:

                if resp.status == 200:
                    result = await resp.json()
                    logger.debug(f"Received Reply {result}")

                    if "errors" in result:
                        raise GraphQLQueryException(
                            f"Ward {self.config.endpoint}:" + str(result["errors"])
                        )

                    return result["data"]

                if resp.status == 403:
                    logger.error("Auth token is expired trying to refresh")
                    raise TokenExpired("Token Expired Error")

                if resp.status == 400:
                    raise GraphQLProtocolException(await resp.json())

                raise GraphQLProtocolException(
                    f"Unexpected statuscode {resp.status} {resp}"
                )

        except TokenExpired as e:
            await self.herre.arefresh()
            return await self.run(gql, variables, retry=retry + 1)

        except aiohttp.client_exceptions.ClientOSError as e:
            await self.adisconnect()
            raise GraphQLProtocolException("Ward is not reachable") from e



from herre.console.context import console, get_current_console
from herre.grants.base import BaseGrant
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from herre.grants.session import OAuth2Session

from herre.grants.registry import register_grant
from herre.config import GrantType


@register_grant(GrantType.CLIENT_CREDENTIALS)
class BackendGrant(BaseGrant):
    refreshable = True

    async def afetch_token(self, **kwargs):
        auth_client = BackendApplicationClient(
            client_id=self.config.client_id, scope=" ".join(self.config.scopes)
        )

        async with OAuth2Session(client=auth_client, scope=self.scope) as session:
            token = await session.fetch_token(
                token_url=self.token_url,
                client_id=self.config.client_id,
                client_secret=self.config.client_secret,
                verify=True,
            )

        return token



from enum import Enum
from typing import List, Optional
import aiohttp

from pydantic.main import BaseModel
import fakts
from herre.config import HerreConfig
from herre.grants.code_server.app import AuthorizationCodeServerGrant
from herre.grants.backend.app import BackendGrant
from herre.grants.backend.app import BackendGrant
import os
import logging
from herre.types import HerreState, User
from koil import get_current_koil, Koil
import time
import shelve

from koil.loop import koil

from fakts import Fakts, get_current_fakts, Config
from herre.grants.registry import GrantRegistry, get_current_grant_registry


logger = logging.getLogger(__name__)


class HerreError(Exception):
    pass


class Herre:
    def __init__(
        self,
        *args,
        register=True,
        config: HerreConfig = None,
        fakts: Fakts = None,
        token_file="token.temp",
        userinfo_url="userinfo/",
        granty_registry: GrantRegistry = None,
        max_retries=5,
        no_temp=False,
        **kwargs,
    ) -> None:
        """Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """
        self.config = config
        self.max_retries = max_retries
        self.fakts: Fakts = fakts or get_current_fakts()
        self.grant_registry = granty_registry or get_current_grant_registry()
        self.no_temp = no_temp
        self.grant = None
        self.state: HerreState = None
        self.token_file = (
            f"{self.fakts.subapp}.{token_file}" if self.fakts.subapp else token_file
        )
        if register:
            set_current_herre(self)

        super().__init__(*args, **kwargs)

    async def alogin(self, force_relogin=False, retry=0, **kwargs) -> HerreState:
        if retry > self.max_retries:
            raise Exception("Exceeded Login Retries")
        if not self.config:
            if not self.fakts.loaded:
                await self.fakts.aload()

            self.config = await HerreConfig.from_fakts(fakts=self.fakts)

        if not force_relogin and not self.config.no_temp and not self.no_temp:
            try:
                with shelve.open(self.token_file) as cfg:
                    client_id = cfg["client_id"]
                    if client_id == self.config.client_id:
                        self.state = HerreState(**cfg["state"])
                    else:
                        logger.info("Omitting old token")

            except Exception:
                pass

        if not self.state:
            self.grant = self.grant_registry.get_grant_for_type(
                self.config.authorization_grant_type
            )(self.config, fakts=self.fakts)
            token_dict = await self.grant.afetch_token(**kwargs)
            logger.info(f"Grant fetched token {token_dict}")
            self.state = HerreState(
                **token_dict, client_id=self.config.client_id, scopes=self.config.scopes
            )

        try:
            base_url = f'{"https" if self.config.secure else "http"}://{self.config.host}:{self.config.port}{self.config.subpath}/'
            user_info_endpoint = base_url + "userinfo/"

            async with aiohttp.ClientSession(
                headers={"Authorization": f"Bearer {self.state.access_token}"}
            ) as session:
                async with session.get(user_info_endpoint) as resp:
                    user_json = await resp.json()
                    if "detail" in user_json:
                        raise Exception(user_json["detail"])

                    try:
                        self.state.user = User(**user_json)
                    except:
                        self.state.user = None

                    if not self.config.no_temp:
                        with shelve.open(self.token_file) as cfg:
                            cfg["client_id"] = self.state.client_id
                            cfg["state"] = self.state.dict()

        except Exception as e:
            logger.error(f"Token Invalid {e}")
            self.state = None
            await self.alogin(force_relogin=True, retry=retry + 1, **kwargs)

        return self.state

    async def alogout(self):
        if self.grant:
            return await self.grant.alogout()

        try:
            with shelve.open(self.token_file) as cfg:
                cfg["state"] = None
                cfg["client_id"] = None
        except KeyError:
            pass

        self.state = None

    async def arefresh(self):
        return await self.grant.alogin()

    def login(self, **kwargs):
        return koil(self.alogin(), **kwargs)

    def logout(self, **kwargs):
        return koil(self.alogout(), **kwargs)

    def refresh(self):
        return koil(self.arefresh())

    @property
    def logged_in(self):
        return self.state is not None

    @property
    def user(self):
        assert self.state, "We are not yet logged in"
        assert self.state.user, "Login is not associated with a user"
        return self.state.user

    @property
    def scopes(self):
        assert self.state, "We are not yet logged in"
        assert self.state.scopes, "Login is not associated with a user"
        return self.state.scopes

    @property
    def headers(self):
        assert self.state, "We are not yet logged in"
        return {"Authorization": f"Bearer {self.state.access_token}"}


CURRENT_HERRE = None


def get_current_herre(**kwargs):
    global CURRENT_HERRE
    if not CURRENT_HERRE:
        CURRENT_HERRE = Herre(**kwargs)
    return CURRENT_HERRE


def set_current_herre(herre):
    global CURRENT_HERRE
    CURRENT_HERRE = herre
